#!/usr/bin/env python
##
## This file is part of Nolimips, a MIPS simulator with unlimited registers
## Copyright (C) 2003, 2004, 2006 Benoit Perrot <benoit@lrde.epita.fr>
##
## Nolimips is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## Nolimips is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##

import sys, os
import string, re

import nolimips
from nolimips import define_id

#### Nolimips' Flex scanner generator --------------------------

if __name__ != "__main__" or len(sys.argv) != 2:
  print "Usage: .. SRCDIR"
  sys.exit(1)
srcdir = sys.argv[1]

parser = nolimips.InstructionSetParser()
instructions = parser.parse(sys.stdin)

sys.stdout = file("asm-scan.ll.tmp", "w")

## Prologue ----------------------------
print "%{                                             /* -*- C++ -*- */"
print nolimips.license
print nolimips.generated_by_hand
print """
#include <string>
#include <sstream>
#include <cerrno>

#include \"common.hh\"
#include \"misc/escape.hh\"

#include \"vm/cpu.hh\"

#include \"parse/asm-scan.hh\"

#define YY_NEVER_INTERACTIVE 1

#define YY_USER_ACTION				\\
    yylloc->columns (yyleng);

using inst::Register;
using vm::Cpu;

typedef yy::parser::token token;

%}

%option noyywrap
%option nounput
%option debug

%x STATE_STRING
eol			\\n|\\r|\\n\\r|\\r\\n

dec_int			[0-9]+
hex_int			0x[0-9A-Fa-f]+

id			[a-zA-Z_][a-zA-Z_0-9]*
%%

%{
  yylloc->step ();
%}

[ \\t]+			yylloc->step ();
[#;][^\\n\\r]*		yylloc->step ();
{eol}			yylloc->lines (1); yylloc->step ();

\"(\"			return token::LPAREN;
\")\"			return token::RPAREN;
\",\"			return token::COMMA;

\"+\"			return token::EXP_ADD;
\"-\"			return token::EXP_SUB;
\"*\"			return token::EXP_MUL;
\"/\"			return token::EXP_DIV;

\".text\"			return token::DIR_TEXT;
\".data\"			return token::DIR_DATA;

\".space\"		return token::DIR_SPACE;
\".byte\"			return token::DIR_BYTE;
\".half\"			return token::DIR_HALF;
\".word\"			return token::DIR_WORD;
\".ascii\"		return token::DIR_ASCII;
\".asciiz\"		return token::DIR_ASCIIZ;
"""

## Body --------------------------------
for inst in instructions:
  print '"' + inst.opcode + \
        '"\t\t\treturn token::' + define_id(inst.opcode) + ";"

## Epilogue ----------------------------
print """
\"$\"([0-9]|[12][0-9]|3[01])	{
  yylval->reg = new Register(Register::generic, atoi(yytext +1));
  return token::REGISTER;
}

\"$zero\"			{
  yylval->reg = new Register(Register::general, Cpu::zero);
  return token::REGISTER;
}"""
for k, r in [("v", range(2)), ("a", range(4)),
             ("t", range(10)), ("s", range(8))]:
  for i in r:
    print "\"$" + k + str(i) + "\"			{"
    print "  yylval->reg = new Register(Register::general, " + \
          "Cpu::" + k + str(i)+ ");"
    print "  return token::REGISTER;\n}"
print """
\"$sp\"			{
  yylval->reg = new Register(Register::general, Cpu::sp);
  return token::REGISTER;
}
\"$fp\"			{
  yylval->reg = new Register(Register::general, Cpu::fp);
  return token::REGISTER;
}
\"$ra\"			{
  yylval->reg = new Register(Register::general, Cpu::ra);
  return token::REGISTER;
}

\"$x\"{dec_int}		{
  yylval->reg = new Register(Register::unlimited, atoi(yytext +2));
  return token::REGISTER;
}


{id}			{
  yylval->id = new misc::unique_string(yytext);
  return token::LABEL;
}
{id}\":\"		{
  yytext[yyleng - 1] = 0; // Remove last character (side-effects on yytext are allowed)
  yylval->id = new misc::unique_string(yytext);
  return token::LABEL_DEF;
}


-?{dec_int}		{
  std::istringstream iss(yytext);

  if (iss.peek() == '-')
    iss >> std::dec >> yylval->i;
  else
    iss >> std::dec >> (unsigned&) yylval->i;

  if (iss.fail())
    {
      std::cerr << *yylloc
		<< \": unrecognized integer (may be too large): \"
		<< yytext << std::endl;
      exit_set (exit_scan);
    }
  return token::INTEGER;
}
-?{hex_int}		{
  std::istringstream iss(yytext);

  if (iss.peek() == '-')
    iss >> std::hex >> yylval->i;
  else
    iss >> std::hex >> (unsigned&) yylval->i;

  if (iss.fail())
    {
      std::cerr << *yylloc
		<< \": unrecognized integer (may be too large): \"
		<< yytext << std::endl;
      exit_set (exit_scan);
    }
  return token::INTEGER;
}

'\\\\a'			yylval->i = '\\a'; return token::INTEGER;
'\\\\b'			yylval->i = '\\b'; return token::INTEGER;
'\\\\f'			yylval->i = '\\f'; return token::INTEGER;
'\\\\n'			yylval->i = '\\n'; return token::INTEGER;
'\\\\r'			yylval->i = '\\r'; return token::INTEGER;
'\\\\t'			yylval->i = '\\t'; return token::INTEGER;
'\\\\v'			yylval->i = '\\v'; return token::INTEGER;
'\\\\[\\\\\\']'		yylval->i = yytext[1]; return token::INTEGER;
'\\\\.'			{
  std::cerr << *yylloc
	    << \": unrecognized escape: \" << escape (yytext) << std::endl;
  exit_set (exit_scan);
}
'[^\\'\\n\\r\\\\]'	yylval->i = yytext[1]; return token::INTEGER;


\\\"			yylval->s = new std::string; BEGIN STATE_STRING;
<STATE_STRING>{
  \\\"			{
    BEGIN INITIAL;
    return token::STRING;
  }

  \\\\[0-7]{3}		{
    long c = strtol (yytext + 1, 0, 8);
    if (c > 255)
      {
	std::cerr << *yylloc
		  << \": invalid escape: \" << escape (yytext) << std::endl;
	exit_set (exit_scan);
      }
    else
      *(yylval->s) += c;
  }
  
  \\\\x[0-9a-fA-F]{2}	{
    *(yylval->s) += strtol (yytext + 2, 0, 16);
  }
  
  \\\\a			*(yylval->s) += '\\a';
  \\\\b			*(yylval->s) += '\\b';
  \\\\f			*(yylval->s) += '\\f';
  \\\\n			*(yylval->s) += '\\n';
  \\\\r			*(yylval->s) += '\\r';
  \\\\t			*(yylval->s) += '\\t';
  \\\\v			*(yylval->s) += '\\v';
  \\\\[\\\\\"\"]		*(yylval->s) += yytext[1];
  \\\\.			{
    std::cerr << *yylloc
	      << \": unrecognized escape: \" << escape (yytext) << std::endl;
    exit_set (exit_scan);
  }
  
  [^\\\"\\n\\r\\\\]+		yylval->s->append (yytext, yyleng);
  
  {eol}			{
    yylloc->lines (1);
    yylval->s->append (\"\\n\", 1);
  }
  
  <<EOF>>		{
    std::cerr
      << *yylloc << \": unexpected end of file in a string\" << std::endl;
    exit_set (exit_scan);
    
    BEGIN INITIAL;
    return token::STRING;
  }
}

.                       {
  std::cerr << *yylloc
	    << \": invalid character: `\"
	    << escape (yytext) << \"'\" << std::endl;
  yylloc->step ();
  exit_set (exit_scan);
}

%%

/* Depending upon the version of Flex (> 2.5.4), there might be a
   function to reclaim all the memory Flex has allocated.  If not,
   simulate it. */
#if !(   YY_FLEX_MAJOR_VERSION >= 2             \
      && YY_FLEX_MINOR_VERSION >= 5             \
      && defined YY_FLEX_SUBMINOR_VERSION       \
      && YY_FLEX_SUBMINOR_VERSION >= 5)
static void
yylex_destroy(void)
{
  yy_delete_buffer(YY_CURRENT_BUFFER);
}
#endif

#include <stack>

namespace parse
{

  // Stack of scanning states, used to have some (weak) reentrancy.
  static std::stack<YY_BUFFER_STATE> states;

  // Keep this scan_open valid for multiple calls (i.e., do use
  // yyrestart).
  void
  scan_open(const std::string &filename, bool trace_p)
  {
    static bool first = true;
    if (first)
      {
        first = false;

        // Reclaim all the memory allocated by Flex at program termination
        std::atexit((void (*) (void)) yylex_destroy);
      }

    // Save the current state
    states.push(YY_CURRENT_BUFFER);
  
    yy_flex_debug = trace_p;

    if (filename == \"-\")
      yyin = stdin;
    else
      {
	yyin = fopen (filename.c_str (), \"r\");
	if (!yyin)
	  {
	    std::cerr
	      << program_name
	      << \": cannot open `\" << filename << \"': \"
	      << strerror (errno) << std::endl;
	    exit (1);
	  }
      }

    yy_switch_to_buffer(yy_create_buffer (yyin, YY_BUF_SIZE));
  }

  void
  scan_close (void)
  {
    fclose (yyin);
    
    // Restore the current scanning state.
    assertion(!states.empty());
    yy_delete_buffer(YY_CURRENT_BUFFER);
    yy_switch_to_buffer(states.top());
    states.pop();
  }
}
"""

sys.stdout.close()
sys.stdout = sys.__stdout__
nolimips.lazy_overwrite(srcdir + os.sep + "asm-scan.ll", "asm-scan.ll.tmp")
os.remove("asm-scan.ll.tmp")
