//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2006 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
#ifndef VM_MMU_HXX
# define VM_MMU_HXX

# include "inst/text_section.hh"

# include "vm/cp0.hh"
# include "vm/memory.hh"

namespace vm
{

  //
  inline
  Mmu::Mmu(Cp0 &cp0, Memory &memory):
    cp0_(cp0),
    memory_(memory),
    text_section_(0)
  {
  }

  inline void
  Mmu::reset()
  {
    // Nothing to do?
  }

  //
  inline bool
  Mmu::translate(int &offset, Memory::segment_kind &k) const
  {
    if (0 <= offset)
      {
	if (offset < memory_.heap_size())
	  {
	    k = Memory::heap;
	    return true;
	  }
	if (memory_.get_stack_top() <= offset)
	  {
	    k = Memory::stack;
	    // Stack grows down
	    offset = offset - memory_.get_stack_top();
	    return true;
	  }
      }
    return false;
  }

  //
  inline void
  Mmu::data_store(const inst::DataSection& data_section)
  {
    memory_.store(data_section);
  }

  inline void
  Mmu::data_store_byte(int offset, int b)
  {
    Memory::segment_kind k;
    if (translate(offset, k))
      memory_.store_byte(k, offset, b);
    else
      cp0_.raise_addr_store();
  }
  inline void
  Mmu::data_store_word(int offset, int w)
  {
    Memory::segment_kind k;
    if (translate(offset, k))
      memory_.store_word(k, offset, w);
    else
      cp0_.raise_addr_store();
  }

  //
  inline int
  Mmu::data_load_byte(int offset) const
  {
    Memory::segment_kind k;
    if (translate(offset, k))
      return memory_.load_byte(k, offset);

    cp0_.raise_addr_load();
    return 0;
  }

  inline int
  Mmu::data_load_word(int offset) const
  {
    Memory::segment_kind k;
    if (translate(offset, k))
      return memory_.load_word(k, offset);

    cp0_.raise_addr_load();
    return 0;
  }

  inline int
  Mmu::data_sbrk(int size)
  {
    return memory_.sbrk(size);
  }

  //
  inline void
  Mmu::inst_store(const inst::TextSection& text_section)
  {
    for (inst::TextSection::instruction_list_type::const_iterator
	   it = text_section.begin(); it != text_section.end(); ++it)
      text_section_.push_back(*it);
  }

  inline const inst::Inst &
  Mmu::inst_load(int offset) const
  {
    // Precondition
    if (offset < 0 ||
	text_section_.size() <= static_cast<unsigned>(offset))
      {
	std::cerr << "inst_load" << std::endl;
	cp0_.raise_addr_load();
	return *text_section_[0]; // FIXME: NO!
      }

    return *text_section_[offset];
  }

} // namespace vm

#endif // !VM_MMU_HXX
